//Author: Vaibhav Rajan, LCBB, EPFL

#include "estimator.h" 
#include "get_tree.h" 
#include <assert.h>
#include <string.h>
#include <unistd.h>
//#include <cstring>
//#include <vector>
//#include <iostream>
//using namespace std;

typedef struct node_struct *node;
typedef struct genome_struct *genome;
typedef struct tree_struct *tree;
typedef struct adj_struct *adjacency;
typedef struct elem_struct *element;

#define MAXNAMESIZE 30

struct node_struct{
  double left_edge_len;
  double right_edge_len;
  int depth;
  node parent;
  node left_child;
  node right_child;
  genome data;
};

struct adj_struct{
  element e1;
  element e2;
};

struct elem_struct{
  adjacency adj;
  int elem_code; //= index; tail: element*2 - 1, head: element*2 
  //int element;
};

struct genome_struct{
  int num_adj;
  adjacency *adjs;//[MAX_ADJ];
  int size_genome;
  element *elems;//[2*NUM_GENES+1]; 
};

struct tree_struct{
  int num_leaves;
  node root;
  node *nodes;
};

/////FUNCTIONS NEEDED TO COMPUTE ESTIMATED TRUE DISTANCE

int num_shared_adj(genome g1, genome g2)
{
  int num = 0;
  int e1, e2;
  for(int i = 0; i < g1->num_adj; i++){
    //printf("is it %d\n", g1->adjs[i] == NULL);
    if(g1->adjs[i]->e1 != NULL)
      e1 = g1->adjs[i]->e1->elem_code;
    else
      printf("error\n");
    if (g2->elems[e1]->adj == NULL) //element deleted in g2
       continue;
    //printf("here\n");
    if(g1->adjs[i]->e2 != NULL){
      e2 = g1->adjs[i]->e2->elem_code;
      //printf("%d %d\n", e1, e2);
      //printf("there %d\n", g2->elems[e1]->adj == NULL);
      if (g2->elems[e2]->adj == NULL)
        continue;
      //printf("continued %d %d\n", e1, e2);
      if ( (e1 <= 2*(g2->size_genome)+1) && (g2->elems[e1]->adj->e1 != NULL) && (g2->elems[e1]->adj->e2 != NULL) && ( (g2->elems[e1]->adj->e1->elem_code == e1 && g2->elems[e1]->adj->e2->elem_code == e2) || (g2->elems[e1]->adj->e1->elem_code == e2 && g2->elems[e1]->adj->e2->elem_code == e1)) ) 
        num++;
      //printf("out\n");
    }
    else if ( (e1 <= 2*(g2->size_genome)+1) && (g2->elems[e1]->adj != NULL) && (g2->elems[e1]->adj->e1 != NULL) && (g2->elems[e1]->adj->e1->elem_code == e1) && (g2->elems[e1]->adj->e2 == NULL) )
        num++;
  }
  return num;
}

int num_lin_chrm(genome g)
{
  int count = 0;
  for(int i = 0; i < g->num_adj; i++){
    if((g->adjs[i]->e2 == NULL) && (g->adjs[i]->e1 != NULL) )
      count++;
  }
  //assert(count%2 == 0);
  return count/2;
}

///////////////////////////////////////////////////////////////
/////DEALLOCATORS

void destroy_genome(genome g)
{
  int i;
  for(i = 0; i < g->num_adj; i++)
    free(g->adjs[i]);
  free(g->adjs);
  for(i = 1; i <= 2*(g->size_genome); i++)
    free(g->elems[i]);
  free(g->elems);
  free(g);
}

void destroy_tree(tree t)
{
  for(int i = 1; i <= t->num_leaves; i++){
    if(t->nodes[i]->data != NULL)
      destroy_genome(t->nodes[i]->data);
    free(t->nodes[i]);
  }
  free(t->nodes);
  free(t);
}

void destroy_leaves(node *leaves, int num_leaves)
{
  for(int i= 1; i <= num_leaves; i++){
    destroy_genome(leaves[i]->data);
    free(leaves[i]);
  }
}

//////////////////////////////////////////////////////////////////
/////COMPUTE AND PRINT DISTANCES TO FILE

void print_all_dist(tree t, FILE* fp, int **dist_mat)
{
  int est;
  fprintf(fp, "%d\n", t->num_leaves);
  for(int i = 1; i <= t->num_leaves; i++){
    for(int j = i+1; j <= t->num_leaves; j++){
      est = (dist_mat[i][j] + dist_mat[j][i])/2;
      fprintf(fp, "%d %d %d\n", i, j, est);
    }
  }
}

void print_all_dist(tree t, FILE *fp)
{
  int sh_adj, l1, l2, est;
  int td;
  int num_genes = t->nodes[1]->data->size_genome;
  fprintf(fp, "%d\n", t->num_leaves);
  for(int i = 1; i <= t->num_leaves; i++){
    for(int j = i+1; j <= t->num_leaves; j++){
      sh_adj = num_shared_adj(t->nodes[i]->data, t->nodes[j]->data);
      l1 = num_lin_chrm(t->nodes[i]->data);
      l2 = num_lin_chrm(t->nodes[j]->data);
      est = GetEstimate(num_genes, l1, l2, sh_adj);
      //est += gaussian(200, 100);
      fprintf(fp, "%d %d %d\n", i, j, est);
      //printf("%d %d %d\n", i, j, est);
    }
  }
}

void print_all_dist(tree t, double **m)
{
  int sh_adj, l1, l2, est;
  int td;
  int num_genes = t->nodes[1]->data->size_genome;
  //fprintf(fp, "%d\n", t->num_leaves);
  for(int i = 1; i <= t->num_leaves; i++){
    for(int j = i+1; j <= t->num_leaves; j++){
      sh_adj = num_shared_adj(t->nodes[i]->data, t->nodes[j]->data);
      l1 = num_lin_chrm(t->nodes[i]->data);
      l2 = num_lin_chrm(t->nodes[j]->data);
      est = GetEstimate(num_genes, l1, l2, sh_adj);
      //fprintf(fp, "%d %d %d\n", i, j, est);
      m[i-1][j-1] = est;
    }
  }
}

void print_pw_dist(node *leaves, int *ext_dist, int num_leaves, int num_genes, FILE *fp)
{
  int sh_adj, l1, l2, est;
  fprintf(fp, "%d\n", num_leaves);
  for(int i = 1; i <= num_leaves; i++){
    for(int j = i+1; j <= num_leaves; j++){
      sh_adj = num_shared_adj(leaves[i]->data, leaves[j]->data);
      l1 = num_lin_chrm(leaves[i]->data);
      l2 = num_lin_chrm(leaves[j]->data);
      est = GetEstimate(num_genes, l1, l2, sh_adj);
      //printf("shared %d %d %d\n", i, j, sh_adj);
      if (ext_dist != NULL)
        est -= (ext_dist[i] + ext_dist[j]);
      //est += gaussian(200, 100);
      fprintf(fp, "%d %d %d\n", i, j, est);
    }
  }
}

void print_pw_dist(node *leaves, int *ext_dist, int num_leaves, int num_genes, double **mlist)
{
  int sh_adj, l1, l2, est;
  //fprintf(fp, "%d\n", num_leaves);
  for(int i = 1; i <= num_leaves; i++){
    for(int j = i+1; j <= num_leaves; j++){
      sh_adj = num_shared_adj(leaves[i]->data, leaves[j]->data);
      l1 = num_lin_chrm(leaves[i]->data);
      l2 = num_lin_chrm(leaves[j]->data);
      est = GetEstimate(num_genes, l1, l2, sh_adj);
      //printf("shared %d %d %d\n", i, j, sh_adj);
      if (ext_dist != NULL)
        est -= (ext_dist[i] + ext_dist[j]);
      //est += gaussian(200, 100);
      //fprintf(fp, "%d %d %d\n", i, j, est);
      mlist[i-1][j-1] = est;
    }
  }
}
/* only for JG
void print_pw_dist(node *leaves, int num_leaves, FILE *fp)
{
  int sh_adj, l1, l2, est;
  fprintf(fp, "%d\n", num_leaves);
  for(int i = 1; i <= num_leaves; i++){
    for(int j = i+1; j <= num_leaves; j++){
      sh_adj = num_shared_adj(leaves[i]->data, leaves[j]->data);
      l1 = num_lin_chrm(leaves[i]->data);
      l2 = num_lin_chrm(leaves[j]->data);
      est = GetEstimate(NUM_GENES, l1, l2, sh_adj);
      //printf("shared %d %d %d\n", i, j, sh_adj);
      fprintf(fp, "%d %d %d\n", i, j, est);
    }
  }
}
*/
/////////////////////////////////////////////////////////////////////

/////PRINT LEAVES: FOR DEBUGGING
void print_leaves(node *leaves, int num_leaves)
{
  for(int i = 1; i <= num_leaves; i++){
    printf("leaf\n");
    for(int j = 0; j < leaves[i]->data->num_adj; j++){
      adjacency a = leaves[i]->data->adjs[j];
      printf("%d ", a->e1->elem_code);
      if (a->e2 != NULL)
        printf("%d | ", a->e2->elem_code);
      else
        printf("NULL | ");
    }
    printf("\n");
  }
  
}


//////////////////////////////////////////////////////////////////////

void copy_leaf(node leaf, node empty)
{
  int e1, e2;
  int num_genes = leaf->data->size_genome;
  genome gen = (genome)malloc(sizeof(struct genome_struct)); 
  gen->adjs = (adjacency *)malloc(2*num_genes*sizeof(adjacency));
  gen->elems = (element *)malloc((2*num_genes+1)*sizeof(element));
  empty->data = gen;
  gen->elems[0] = NULL;
  for(int i = 1; i < (2*(leaf->data->size_genome) + 1); i++){
    gen->elems[i] = (element)malloc(sizeof(struct elem_struct));
    gen->elems[i]->elem_code = i;
  }
  for(int i = 0; i < leaf->data->num_adj; i++){
    gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
    e1 = leaf->data->adjs[i]->e1->elem_code;
    gen->adjs[i]->e1 = gen->elems[e1];
    gen->elems[e1]->adj = gen->adjs[i];
    if (leaf->data->adjs[i]->e2 != NULL){ 
      e2 = leaf->data->adjs[i]->e2->elem_code;
      gen->adjs[i]->e2 = gen->elems[e2];
      gen->elems[e2]->adj = gen->adjs[i];
    }
    else
      gen->adjs[i]->e2 = NULL;
  }
  gen->size_genome = leaf->data->size_genome;
  gen->num_adj = leaf->data->num_adj;
}    

/////BOOTSTRAP BY PERTURBATIONS
//BP

void evolveDCJ(node parent, node child, int num_times, int num_genes)
{
  genome pgen = parent->data;
  genome gen = child->data;

  int cut1, cut2, num_changes, e1, e2, e3, e4; 
  //printf("evolving dcj %d times\n", num_times);
  while(num_times--){
    //printf("dcj %d adj %d\n", num_times, pgen->num_adj);
    while (pgen->num_adj <= (cut1 = rand() / (RAND_MAX/pgen->num_adj)));
    while (pgen->num_adj <= (cut2 = rand() / (RAND_MAX/pgen->num_adj)));

    //printf("%d %d\n", cut1, cut2);
    /*
    if (cut1 != cut2 && pgen->adjs[cut1]->e2 == NULL && pgen->adjs[cut2]->e2 == NULL)
      num_changes = -1;
    else if (cut1 == cut2 && pgen->adjs[cut1]->e2 != NULL && pgen->adjs[cut2]->e2 != NULL)//same adj
      num_changes = 1;
    else
      num_changes = 0;
    */
    if(child->data == NULL){
      //printf("first time\n");
      gen = (genome)malloc(sizeof(struct genome_struct)); 
      gen->adjs = (adjacency *)malloc((2*num_genes)*sizeof(adjacency));
      gen->elems = (element *)malloc((2*num_genes+1)*sizeof(element));
      child->data = gen;
      gen->elems[0] = NULL;
      for(int i = 1; i < (2*(pgen->size_genome) + 1); i++){
        gen->elems[i] = (element)malloc(sizeof(struct elem_struct));
        gen->elems[i]->elem_code = i;
      }
      for(int i = 0; i < pgen->num_adj; i++){
        gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
        e1 = pgen->adjs[i]->e1->elem_code;
        gen->adjs[i]->e1 = gen->elems[e1];
        gen->elems[e1]->adj = gen->adjs[i];
        if (pgen->adjs[i]->e2 != NULL){ 
          e2 = pgen->adjs[i]->e2->elem_code;
          gen->adjs[i]->e2 = gen->elems[e2];
          gen->elems[e2]->adj = gen->adjs[i];
        }
        else
          gen->adjs[i]->e2 = NULL;
      }
      gen->size_genome = pgen->size_genome;
      gen->num_adj = pgen->num_adj;
    }
    //gen->size_genome = pgen->size_genome;
    //gen->num_adj = pgen->num_adj + num_changes;
    //gen->adjs = (adjacency *)malloc(sizeof(adjacency) * gen->num_adj);
    //gen->elems = (element *)malloc(sizeof(element) * (2*(gen->size_genome) + 1));
   
    if (cut1 == cut2 && pgen->adjs[cut1]->e2 != NULL && pgen->adjs[cut2]->e2 != NULL){
      e1 = pgen->adjs[cut1]->e1->elem_code;
      e2 = pgen->adjs[cut1]->e2->elem_code;
      //printf("%d %d\n", e1, e2);
      gen->adjs[cut1]->e1 = gen->elems[e1];//necessary?
      gen->elems[e1]->adj = gen->adjs[cut1];
      gen->adjs[cut1]->e2 = NULL;
      int i = gen->num_adj;
      //if (gen->adjs[i] == NULL)
        gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[i]->e1 = gen->elems[e2];
      gen->elems[e2]->adj = gen->adjs[i];
      gen->adjs[i]->e2 = NULL;
      gen->num_adj += 1;
    }
    else if (cut1 != cut2 && pgen->adjs[cut1]->e2 == NULL && pgen->adjs[cut2]->e2 == NULL){
      e1 = pgen->adjs[cut1]->e1->elem_code;
      e2 = pgen->adjs[cut2]->e1->elem_code;
      //printf("%d %d\n", e1, e2);
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut1]->e1 = gen->elems[e1];
      gen->elems[e1]->adj = gen->adjs[cut1];
      gen->adjs[cut1]->e2 = gen->elems[e2];
      gen->elems[e2]->adj = gen->adjs[cut1];
      int i = gen->num_adj - 1;
      adjacency tmp = gen->adjs[cut2];
      gen->adjs[cut2] = gen->adjs[i];
      free(tmp);
      gen->adjs[i] = NULL;
      gen->num_adj -= 1;
    }
    else if (cut1 != cut2 && pgen->adjs[cut1]->e2 == NULL){
      if (rand() < RAND_MAX/2){
        e1 = pgen->adjs[cut2]->e1->elem_code;
        e2 = pgen->adjs[cut2]->e2->elem_code;
      }
      else{
        e2 = pgen->adjs[cut2]->e1->elem_code;
        e1 = pgen->adjs[cut2]->e2->elem_code;
      }
      //printf("%d %d\n", e1, e2);
      e3 = pgen->adjs[cut1]->e1->elem_code;
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut2]->e1 = gen->elems[e1];
      gen->elems[e1]->adj = gen->adjs[cut2];
      gen->adjs[cut2]->e2 = gen->elems[e3];
      gen->elems[e3]->adj = gen->adjs[cut2];
    
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut1]->e1 = gen->elems[e2];
      gen->elems[e2]->adj = gen->adjs[cut1];
      gen->adjs[cut1]->e2 = NULL;
    }
    else if (cut1 != cut2 && pgen->adjs[cut2]->e2 == NULL){
      if (rand() < RAND_MAX/2){
        e1 = pgen->adjs[cut1]->e1->elem_code;
        e2 = pgen->adjs[cut1]->e2->elem_code;
      }
      else{
        e2 = pgen->adjs[cut1]->e1->elem_code;
        e1 = pgen->adjs[cut1]->e2->elem_code;
      }
      //printf("%d %d\n", e1, e2);
      e3 = pgen->adjs[cut2]->e1->elem_code;
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut1]->e1 = gen->elems[e1];
      gen->elems[e1]->adj = gen->adjs[cut1];
      gen->adjs[cut1]->e2 = gen->elems[e3];
      gen->elems[e3]->adj = gen->adjs[cut1];
    
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut2]->e1 = gen->elems[e2];
      gen->elems[e2]->adj = gen->adjs[cut2];
      gen->adjs[cut2]->e2 = NULL;
    }  
    else if (cut1 != cut2){
      if (rand() < RAND_MAX/2){
        e2 = pgen->adjs[cut2]->e1->elem_code;
        e4 = pgen->adjs[cut2]->e2->elem_code;
      }
      else{
        e4 = pgen->adjs[cut2]->e1->elem_code;
        e2 = pgen->adjs[cut2]->e2->elem_code;
      }
      e1 = pgen->adjs[cut1]->e1->elem_code;
      e3 = pgen->adjs[cut1]->e2->elem_code;
      //printf("%d %d\n", e1, e2);
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut1]->e1 = gen->elems[e1];
      gen->elems[e1]->adj = gen->adjs[cut1];
      gen->adjs[cut1]->e2 = gen->elems[e2];
      gen->elems[e2]->adj = gen->adjs[cut1];
    
      //gen->adjs[i] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[cut2]->e1 = gen->elems[e3];
      gen->elems[e3]->adj = gen->adjs[cut2];
      gen->adjs[cut2]->e2 = gen->elems[e4];
      gen->elems[e4]->adj = gen->adjs[cut2];
    }
    //else 
      //printf("WTF! cut1 %d cut2 %d\n", cut1, cut2);
    //getchar();
    pgen = gen;
  }
}

double gaussian(double mean, double sd)
{
  static double t = 0.0;
  double x,v1,v2,r;
  if (t == 0) {
    do {
      v1 = 2.0 * (rand()/((double)RAND_MAX+1)) - 1.0;
      v2 = 2.0 * (rand()/((double)RAND_MAX+1)) - 1.0;
      r = v1 * v1 + v2 * v2;
    } while (r>=1.0 || r == 0.0);
    r = sqrt((-2.0*log(r))/r);
    t = v2*r;
    return(mean+v1*r*sd);
  }
  else {
    x = t;
    t = 0.0;
    return(mean+x*sd);
  }
}

node *extend_evolution(node *leaves, int *ext_dist, int num_leaves, double mean, double sd, int num_genes)
{
  int gauss;
  for(int i = 1; i <= num_leaves; i++){
     while ((gauss = (int)gaussian(mean, sd)) < 0);
     //printf("extension %d\n", gauss);
     //getchar();
     ext_dist[i] = gauss;
     evolveDCJ(leaves[i], leaves[i], gauss, num_genes);
  }
  return leaves;
}

void run_bp(char *filename, tree t, int num_bs)
{
  int b, i;
  FILE *fq = fopen(filename, "w");
  int num_genes = t->nodes[1]->data->size_genome;
  print_all_dist(t,fq);
  printf("Computing bootstrap support values\n");
  //print_leaves(t->nodes, t->num_leaves); 
  node *leaves;
  int *ext_dist;
  leaves = (node *)malloc(sizeof(node) * (1 + t->num_leaves));
  ext_dist = (int *)malloc(sizeof(int) * (1 + t->num_leaves));
  leaves[0] = NULL;
  printf("Computing bootstrap support values\n");
  for(b = 0; b < num_bs; b++){
    for(i = 1; i <= t->num_leaves; i++){
      leaves[i] = (node)malloc(sizeof(node_struct));
      copy_leaf(t->nodes[i], leaves[i]);
    }
    //printf("extend evolution %d\n", b);
    leaves = extend_evolution(leaves, ext_dist, t->num_leaves, 0.1*num_genes, 0.04*num_genes, num_genes);
    //print_leaves(leaves, t->num_leaves);
    //for (int i = 1; i <=t->num_leaves; i++)
      //printf ("%d ", ext_dist[i]);
    //printf("extension complete\n");
    //getchar();
    print_pw_dist(leaves, ext_dist, t->num_leaves, num_genes, fq);
    destroy_leaves(leaves, t->num_leaves);
  }
  free(leaves);
  free(ext_dist);  
  fclose(fq);
}

void run_bp(tree t, int num_bs, double ***mlist)
{
  int b, i;
  //FILE *fq = fopen(filename, "w");
  int num_genes = t->nodes[1]->data->size_genome;
  print_all_dist(t,mlist[0]);
  //printf("writing to %s\n", filename);
  //print_leaves(t->nodes, t->num_leaves); 
  node *leaves;
  int *ext_dist;
  leaves = (node *)malloc(sizeof(node) * (1 + t->num_leaves));
  ext_dist = (int *)malloc(sizeof(int) * (1 + t->num_leaves));
  leaves[0] = NULL;
  printf("Computing bootstrap support values\n");
  for(b = 1; b <= num_bs; b++){
    for(i = 1; i <= t->num_leaves; i++){
      leaves[i] = (node)malloc(sizeof(node_struct));
      copy_leaf(t->nodes[i], leaves[i]);
    }
    //printf("extend evolution %d\n", b);
    leaves = extend_evolution(leaves, ext_dist, t->num_leaves, 0.1*num_genes, 0.04*num_genes, num_genes);
    //print_leaves(leaves, t->num_leaves);
    //for (int i = 1; i <=t->num_leaves; i++)
      //printf ("%d ", ext_dist[i]);
    //printf("extension complete\n");
    //getchar();
    print_pw_dist(leaves, ext_dist, t->num_leaves, num_genes, mlist[b]);
    destroy_leaves(leaves, t->num_leaves);
  }
  free(leaves);
  free(ext_dist);  
  //fclose(fq);
}


//BP

/////JACKKNIFE GENES
//JG
/* retain for experiments
node *jackknife(node *leaves, int num_leaves, int num_drop)
{
  int gene_list[NUM_GENES] = {0};
  int *del_list = (int *)malloc(sizeof(int) * num_drop);
  int num_marked = 0;
  int sel;
  while(num_marked != num_drop){
    while (NUM_GENES <= (sel = rand() / (RAND_MAX/NUM_GENES)));
    if (gene_list[sel] == 0){
      gene_list[sel] = 1;
      num_marked++;
    }
  }
  int k = 0;
  for(int j = 0; j < NUM_GENES; j++){
     if(gene_list[j] == 1){
       del_list[k] = j+1;
       k++;
     }
  }
  assert(k == num_drop);
  //printf("found genes to delete\n");
  for(int i = 1; i <= num_leaves; i++){
    genome g = leaves[i]->data;
//for(int c = 0; c < g->num_adj; c++)
//  assert(g->adjs[c]->e1 != NULL);
//for(int c = 1; c <= 2*NUM_GENES; c++)
//  assert(g->elems[c]->adj->e1 != NULL);
    for(int j = 0; j < num_drop; j++){
      int elem = del_list[j];
      //printf("i %d j %d elem %d %d\n", i,j, elem*2, elem*2-1);
      element head = g->elems[elem*2];
      element tail = g->elems[elem*2 - 1];
      //printf("elem codes %d %d\n", head->elem_code, tail->elem_code);
      assert(head->adj != NULL);
      assert(tail->adj != NULL);
      assert(head->adj->e1 != NULL);
      assert(tail->adj->e1 != NULL);
//getchar();
      if (head->adj == tail->adj){//del adj
        //printf("if\n");
        int l = g->num_adj - 1;
        tail->adj->e1 = g->adjs[l]->e1;
        tail->adj->e2 = g->adjs[l]->e2;
        g->adjs[l]->e1->adj = tail->adj;
        if (g->adjs[l]->e2 != NULL)
          g->adjs[l]->e2->adj = tail->adj;
        free(g->adjs[l]);
        g->num_adj -= 1;
        head->adj = NULL;
        tail->adj = NULL;
      }
      else{
        //printf("else\n");
        element head_other, tail_other;
        //printf("%d %d\n", head->adj->e1->elem_code, tail->adj->e1->elem_code);
        //printf("%d %d\n", head->adj->e2==NULL, tail->adj->e2==NULL);
//getchar();
        if(head->adj->e1 == head)
          head_other = head->adj->e2;
        else
          head_other = head->adj->e1;
        if(tail->adj->e1 == tail)
          tail_other = tail->adj->e2;
        else
          tail_other = tail->adj->e1;
        //printf("%d %d\n", head_other==NULL, tail_other==NULL);
        if (head_other == NULL && tail_other != NULL){
          head->adj->e2 = NULL;
          head->adj->e1 = tail_other;
          tail_other->adj = head->adj;
        }
        else if (head_other != NULL && tail_other == NULL){
          head->adj->e1 = head_other;
          head->adj->e2 = NULL;
          head_other->adj = head->adj;
        }
        else if (head_other != NULL && tail_other != NULL){
          head->adj->e1 = head_other;
          head->adj->e2 = tail_other;
          head_other->adj = head->adj;
          tail_other->adj = head->adj;
        }
        else {//both null
          int l = g->num_adj - 1;
          head->adj->e1 = g->adjs[l]->e1;
          head->adj->e2 = g->adjs[l]->e2;
          g->adjs[l]->e1->adj = head->adj;
          if (g->adjs[l]->e2 != NULL)
            g->adjs[l]->e2->adj = head->adj;
          free(g->adjs[l]);
          g->num_adj -= 1;
        }
        //printf("deleting\n");
        int l = g->num_adj - 1;
        //printf("last %d\n", l);
        
        if(g->adjs[l] != tail->adj){
          tail->adj->e1 = g->adjs[l]->e1;
          //printf("here1\n");
          tail->adj->e2 = g->adjs[l]->e2;
          //printf("here2 %d\n", g->adjs[l]->e1 == NULL);
          if (g->adjs[l]->e1 != NULL)
            g->adjs[l]->e1->adj = tail->adj;
          //printf("here3\n");
          if (g->adjs[l]->e2 != NULL)
            g->adjs[l]->e2->adj = tail->adj;
          //printf("here4\n");
        }
        free(g->adjs[l]);
        g->adjs[l] = NULL;
        g->num_adj -= 1;
        head->adj = NULL;
        tail->adj = NULL;
        //printf("end\n");
      } 
    }
  }
  free(del_list);
  return leaves;
}

void run_jg(char *filename, tree t)
{
  #define JK_PERC 0.4
  int b, i;
  FILE *fj = fopen(filename, "w"); //jackknife jijun
  print_all_dist(t,fj);
  printf("writing to %s\n", filename);
  //print_leaves(t->nodes, t->num_leaves); 
  node *leaves;
  //int *ext_dist;
  leaves = (node *)malloc(sizeof(node) * (1 + t->num_leaves));
  leaves[0] = NULL;
  for(b = 0; b < NUM_BOOTSTRAPS; b++){
    for(i = 1; i <= t->num_leaves; i++){
      leaves[i] = (node)malloc(sizeof(node_struct));
      copy_leaf(t->nodes[i], leaves[i]);
    }
    printf("jackknife %d\n", b);
    //leaves = extend_evolution(leaves, ext_dist, t->num_leaves, 0.1*NUM_GENES, 0.04*NUM_GENES);
    leaves = jackknife(leaves, t->num_leaves, JK_PERC*NUM_GENES);
    //print_leaves(leaves, t->num_leaves);
    //for (int i = 1; i <=t->num_leaves; i++)
      //printf ("%d ", ext_dist[i]);
    printf("jackknife complete\n");
    //getchar();
    print_pw_dist(leaves, t->num_leaves, fj);
    destroy_leaves(leaves, t->num_leaves);
  }
  free(leaves);
  //free(ext_dist);
  fclose(fj);
}
*/
//JG


/////BOOTSTRAP BY PAIRWISE COMPARISONS
//BPC
/* retain for experiments
int get_shared_adj(node l1, node l2, vector<int> &sel_adj)
{
  int e1, ind;
  int count = 0;
  adjacency adj2;
  for(int i = 0; i < sel_adj.size(); i++){
    ind = sel_adj[i];
    e1 = l1->data->adjs[ind]->e1->elem_code;
    adj2 = l2->data->elems[e1]->adj;
    if(adj2->e1->elem_code == e1){
      if(adj2->e2 != NULL && l1->data->adjs[ind]->e2 != NULL){
        if(adj2->e2->elem_code == l1->data->adjs[ind]->e2->elem_code)
          count++;
      }
      else if (adj2->e2 == NULL && l1->data->adjs[ind]->e2 == NULL)
        count++;
    }
    else if(l1->data->adjs[ind]->e2 != NULL && adj2->e1->elem_code == l1->data->adjs[ind]->e2->elem_code){
      if(adj2->e2 != NULL){
        if(adj2->e2->elem_code == e1)
          count++;
      }
    }
  }
  return count;
}

void get_selected_adj(node leaf, vector<int> &sel_adj)
{
  int ind, num = leaf->data->num_adj;
  for(int i = 0; i < num; i++){
    while (num <= (ind = (int)((double)rand() / ((double)RAND_MAX/num))));
    sel_adj.push_back(ind);
  }
}

void run_bpc(char *filename, tree t)
{
  int i, j, b, l1, l2, est, sh_adj;
  FILE *fr = fopen(filename, "w");
  print_all_dist(t,fr);
  printf("writing to %s\n", filename);
  int **dist_mat = (int **)malloc(sizeof(int *) * (1+t->num_leaves));
  for(i = 1; i <= t->num_leaves; i++)
    dist_mat[i] = (int *)calloc(1+t->num_leaves, sizeof(int));
  vector<int> selected_adj;
  for(b = 0; b < NUM_BOOTSTRAPS; b++){
printf("bs %d\n", b);
    for (i = 1; i <= t->num_leaves; i++){
      selected_adj.clear();
      get_selected_adj(t->nodes[i], selected_adj);
      for (j = 1; j <= t->num_leaves; j++){
        if (i != j){
          sh_adj = get_shared_adj(t->nodes[i], t->nodes[j], selected_adj);
          l1 = num_lin_chrm(t->nodes[i]->data);
          l2 = num_lin_chrm(t->nodes[j]->data);
          est = GetEstimate(NUM_GENES, l1, l2, sh_adj);
          if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
            est = 0;
          dist_mat[i][j] = est;
        }
      }
    }
    print_all_dist(t,fr, dist_mat);
  }
  fclose(fr);
  for(i = 1; i <= t->num_leaves; i++)
    free(dist_mat[i]);
  free(dist_mat);
}
*/
//BPC

//////////////////////////////////////////////////////////////////////
void unset_adjacencies(int **sel_adj, int num_genes)
{
  int num =(2*num_genes+1);
  for(int i = 0; i < num; i++){
    for(int j = 0; j < num; j++)
      sel_adj[i][j] = 0;
  }
}

int get_shared_adj(node l1, node l2, int **sel_adj)
{
  int e1;
  int count = 0;
  adjacency adj2;
  for(int i = 0; i < l1->data->num_adj; i++){
    e1 = l1->data->adjs[i]->e1->elem_code;
    adj2 = l2->data->elems[e1]->adj;
    if(adj2->e1->elem_code == e1){
      if(adj2->e2 != NULL && l1->data->adjs[i]->e2 != NULL){
        if(adj2->e2->elem_code == l1->data->adjs[i]->e2->elem_code){
          count += sel_adj[e1][adj2->e2->elem_code];
          count += sel_adj[adj2->e2->elem_code][e1];
        }
      }
      else if (adj2->e2 == NULL && l1->data->adjs[i]->e2 == NULL){
        count += sel_adj[0][e1];
      }
    }
    else if(l1->data->adjs[i]->e2 != NULL && adj2->e1->elem_code == l1->data->adjs[i]->e2->elem_code){
      if(adj2->e2 != NULL){
        if(adj2->e2->elem_code == e1){
          count += sel_adj[adj2->e2->elem_code][adj2->e1->elem_code];
          count += sel_adj[adj2->e1->elem_code][adj2->e2->elem_code];
        }
      }
    }
  }
  return count; 
}

/////PARSIMONY JACKKNIFE
//PJ

void set_adjacencies(int **sel_adj, int num_genes)
{
  int ind1, ind2;
  int num = 2*num_genes+1;
  for(int i = 0; i < (num*(num-1))/2; i++){
    while (num <= (ind1 = (int)((double)rand() / ((double)RAND_MAX/num))));
    while (num <= (ind2 = (int)((double)rand() / ((double)RAND_MAX/num))));
    if (ind1 == ind2){
      i--;
      continue;
    }
    else if (ind1 < ind2){
      if (sel_adj[ind1][ind2] == 0)
        (sel_adj[ind1][ind2])++;
    }
    else{
      if (sel_adj[ind2][ind1] == 0)
        (sel_adj[ind2][ind1])++;
    }
  }

}

void run_pj(char *filename, tree t, int num_bs)
{
  int i, j, b, l1, l2, est, sh_adj;
  int num_genes = t->nodes[1]->data->size_genome;
  int **sel_adj = (int **)calloc(2*num_genes+1,sizeof(int *));
  for(i = 0; i < 2*num_genes + 1; i++)
    sel_adj[i] = (int *)calloc(2*num_genes+1, sizeof(int));

  FILE *fs = fopen(filename, "w");
  print_all_dist(t,fs);
  printf("Computing bootstrap support values\n");
  for(b = 0; b < num_bs; b++){
    set_adjacencies(sel_adj, num_genes);
    //printf("bs %d\n", b);
    fprintf(fs, "%d\n", t->num_leaves);
    for (i = 1; i <= t->num_leaves; i++){
      for (j = i+1; j <= t->num_leaves; j++){
        sh_adj = get_shared_adj(t->nodes[i], t->nodes[j], sel_adj);
        l1 = num_lin_chrm(t->nodes[i]->data);
        l2 = num_lin_chrm(t->nodes[j]->data);
        est = GetEstimate(num_genes, l1, l2, sh_adj);
        if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
          est = 0;
        fprintf(fs, "%d %d %d\n", i, j, est);
      }  
    }
    //memset(sel_adj, 0, (2*NUM_GENES+1)*(2*NUM_GENES+1));
    unset_adjacencies(sel_adj, num_genes);
  }
  fclose(fs);
  for(i = 0; i < 2*num_genes + 1; i++)
    free(sel_adj[i]);
  free(sel_adj);  
}

void run_pj(tree t, int num_bs, double ***mlist)
{
  int i, j, b, l1, l2, est, sh_adj;
  int num_genes = t->nodes[1]->data->size_genome;
  int **sel_adj = (int **)calloc(2*num_genes+1,sizeof(int *));
  for(i = 0; i < 2*num_genes + 1; i++)
    sel_adj[i] = (int *)calloc(2*num_genes+1, sizeof(int));

  //FILE *fs = fopen(filename, "w");
  print_all_dist(t, mlist[0]);
  //printf("writing to %s\n", filename);
  printf("Computing bootstrap support values\n");
  for(b = 1; b <= num_bs; b++){
    set_adjacencies(sel_adj, num_genes);
    //printf("bs %d\n", b);
    //fprintf(fs, "%d\n", t->num_leaves);
    for (i = 1; i <= t->num_leaves; i++){
      for (j = i+1; j <= t->num_leaves; j++){
        sh_adj = get_shared_adj(t->nodes[i], t->nodes[j], sel_adj);
        l1 = num_lin_chrm(t->nodes[i]->data);
        l2 = num_lin_chrm(t->nodes[j]->data);
        est = GetEstimate(num_genes, l1, l2, sh_adj);
        if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
          est = 0;
        mlist[b][i-1][j-1] = est;
        //fprintf(fs, "%d %d %d\n", i, j, est);
      }  
    }
    //memset(sel_adj, 0, (2*NUM_GENES+1)*(2*NUM_GENES+1));
    unset_adjacencies(sel_adj, num_genes);
  }
  //fclose(fs);
  for(i = 0; i < 2*num_genes + 1; i++)
    free(sel_adj[i]);
  free(sel_adj);  
}

//PJ

/////BOOTSTRAP CLASSIC
//BC


void set_adjacencies_oc(int **sel_adj, int num_genes)
{
  int ind1, ind2;
  int num = 2*num_genes+1;

  for(int i = 0; i < (num*(num-1))/2; i++){
    while (num <= (ind1 = (int)((double)rand() / ((double)RAND_MAX/num))));
    while (num <= (ind2 = (int)((double)rand() / ((double)RAND_MAX/num))));
    assert(ind1 < num);
    assert(ind2 < num);
    if (ind1 == ind2){
      i--;
      continue;
    }
    else if (ind1 < ind2){
      (sel_adj[ind1][ind2])++;
    }
    else{
      (sel_adj[ind2][ind1])++;
    }
  }

}

void run_bc(char *filename, tree t, int num_bs)
{
  int i, j, b, l1, l2, est, sh_adj;
  int num_genes = t->nodes[1]->data->size_genome;
  int **sel_adj = (int **)calloc(2*num_genes+1,sizeof(int *));
  for(i = 0; i < 2*num_genes + 1; i++)
    sel_adj[i] = (int *)calloc(2*num_genes+1, sizeof(int));

  FILE *fp = fopen(filename, "w");
  print_all_dist(t,fp);
  printf("Computing bootstrap support values\n");
  for(b = 0; b < num_bs; b++){
    set_adjacencies_oc(sel_adj, num_genes);
    //printf("bs %d\n", b);
    fprintf(fp, "%d\n", t->num_leaves);
    for (i = 1; i <= t->num_leaves; i++){
      for (j = i+1; j <= t->num_leaves; j++){
        //printf("\ti j %d %d\t", i, j);
        sh_adj = get_shared_adj(t->nodes[i], t->nodes[j], sel_adj);
        l1 = num_lin_chrm(t->nodes[i]->data);
        l2 = num_lin_chrm(t->nodes[j]->data);
        est = GetEstimate(num_genes, l1, l2, sh_adj);
        if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
          est = 0;
        //printf("\t%d sh adj %d %d %d\n", est, sh_adj, l1, l2);
        fprintf(fp, "%d %d %d\n", i, j, est);
      }  
    }
    //memset(sel_adj, 0, (2*NUM_GENES+1)*(2*NUM_GENES+1));
    unset_adjacencies(sel_adj, num_genes);
  }
  fclose(fp);

  for(i = 0; i < 2*num_genes + 1; i++)
    free(sel_adj[i]);
  free(sel_adj);  
}

void run_bc(tree t, int num_bs, double ***mlist)
{
  int i, j, b, l1, l2, est, sh_adj;
  int num_genes = t->nodes[1]->data->size_genome;
  int **sel_adj = (int **)calloc(2*num_genes+1,sizeof(int *));
  for(i = 0; i < 2*num_genes + 1; i++)
    sel_adj[i] = (int *)calloc(2*num_genes+1, sizeof(int));

  //FILE *fp = fopen(filename, "w");
  print_all_dist(t,mlist[0]);
  //printf("writing to %s\n", filename);
  printf("Computing bootstrap support values\n");
  for(b = 1; b <= num_bs; b++){
    set_adjacencies_oc(sel_adj, num_genes);
    //printf("bs %d\n", b);
    //fprintf(fp, "%d\n", t->num_leaves);
    for (i = 1; i <= t->num_leaves; i++){
      for (j = i+1; j <= t->num_leaves; j++){
        sh_adj = get_shared_adj(t->nodes[i], t->nodes[j], sel_adj);
        //printf("sh adj %d %d %d\n", sh_adj, l1, l2);
        l1 = num_lin_chrm(t->nodes[i]->data);
        l2 = num_lin_chrm(t->nodes[j]->data);
        est = GetEstimate(num_genes, l1, l2, sh_adj);
        if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
          est = 0;
        //fprintf(fp, "%d %d %d\n", i, j, est);
        mlist[b][i-1][j-1] = est;
      }  
    }
    //memset(sel_adj, 0, (2*NUM_GENES+1)*(2*NUM_GENES+1));
    unset_adjacencies(sel_adj, num_genes);
  }
  //fclose(fp);

  for(i = 0; i < 2*num_genes + 1; i++)
    free(sel_adj[i]);
  free(sel_adj);  
}


//BC


/////JACKKNIFE
//JK

int get_shared_adj_jk(node l1, node l2, int **sel_adj)
{
  int e1, a1, a2;
  int count = 0;
  adjacency adj2;
  for(int i = 0; i < l1->data->num_adj; i++){
    e1 = l1->data->adjs[i]->e1->elem_code;
    adj2 = l2->data->elems[e1]->adj;
    if(adj2->e1->elem_code == e1){
      if(adj2->e2 != NULL && l1->data->adjs[i]->e2 != NULL){
        if(adj2->e2->elem_code == l1->data->adjs[i]->e2->elem_code){
          if (e1 < adj2->e2->elem_code){
            a1 = e1;
            a2 = adj2->e2->elem_code;
          }
          else{
            a1 = adj2->e2->elem_code;
            a2 = e1;
          }
          if(sel_adj[a1][a2] != -1)
            count += sel_adj[a1][a2];
        }
      }
      else if (adj2->e2 == NULL && l1->data->adjs[i]->e2 == NULL){
        if(sel_adj[0][e1] != -1)
          count += sel_adj[0][e1];
      }
    }
    else if(l1->data->adjs[i]->e2 != NULL && adj2->e1->elem_code == l1->data->adjs[i]->e2->elem_code){
      if(adj2->e2 != NULL){
        if(adj2->e2->elem_code == e1){
          if (e1 < adj2->e1->elem_code){
            a1 = e1;
            a2 = adj2->e1->elem_code;
          }
          else{
            a1 = adj2->e1->elem_code;
            a2 = e1;
          }
          if(sel_adj[a1][a2] != -1)
            count += sel_adj[a1][a2];
        }
      }
    }
  }
  return count;
}

void jackknife(int **sel_adj, int num, int num_genes)
{
  int size = 2*num_genes + 1;
  int ind1, ind2, tmp;
  while(num != 0){
    while (size <= (ind1 = (int)((double)rand() / ((double)RAND_MAX/size))));
    while (size <= (ind2 = (int)((double)rand() / ((double)RAND_MAX/size))));
    //printf("%d %d\n", ind1, ind2);
    if(ind1 > ind2){
      tmp = ind2;
      ind2 = ind1;
      ind1 = tmp;
    }
    if (sel_adj[ind1][ind2] == 1){
      sel_adj[ind1][ind2] = -1;
      num--;
    }
  }
}

int set_adj_jk(int **sel_adj, node l)
{
  int e1, e2, tmp;
  adjacency adj;
  int num = 0;
  for(int i = 0; i < l->data->num_adj; i++){
    adj = l->data->adjs[i];
    e1 = adj->e1->elem_code;
    if(adj->e2 != NULL){
      e2 = adj->e2->elem_code;
      if(e1 > e2){
        tmp = e2;
        e2 = e1;
        e1 = tmp;
      }
      if (sel_adj[e1][e2] != 1){
        sel_adj[e1][e2] = 1; 
        num++;
      }
    }
    else{
      if (sel_adj[0][e1] != 1){
        sel_adj[0][e1] = 1;
        num++;
      }
    }
  }
  return num;
}

void run_jk(char *filename, tree t, double perc, int num_bs)
{
  int i, j, b, l1, l2, est, sh_adj;
  int num_genes = t->nodes[1]->data->size_genome;
  int **sel_adj = (int **)calloc(2*num_genes+1,sizeof(int *));
  for(i = 0; i < 2*num_genes + 1; i++)
    sel_adj[i] = (int *)calloc(2*num_genes+1, sizeof(int));

  FILE *fj2 = fopen(filename, "w"); //jackknife
  print_all_dist(t,fj2);
  printf("Computing bootstrap support values\n");
  for(b = 0; b < num_bs; b++){
    //printf("bs %d\n", b);
    int num_adj = 0;
    unset_adjacencies(sel_adj, num_genes);
    for(i = 1; i <= t->num_leaves; i++)
      num_adj += set_adj_jk(sel_adj, t->nodes[i]);
    jackknife(sel_adj, int(double(num_adj)*perc), num_genes);
    fprintf(fj2, "%d\n", t->num_leaves);
    for (i = 1; i <= t->num_leaves; i++){
      for (j = i+1; j <= t->num_leaves; j++){
        sh_adj = get_shared_adj_jk(t->nodes[i], t->nodes[j], sel_adj);
        l1 = num_lin_chrm(t->nodes[i]->data);
        l2 = num_lin_chrm(t->nodes[j]->data);
//printf("sh adj %d %d %d\n", sh_adj, l1, l2);
        est = GetEstimate(num_genes, l1, l2, sh_adj);
        if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
          est = 0;
        fprintf(fj2, "%d %d %d\n", i, j, est);
        //printf("sh adj %d %d %d %d %d\n", sh_adj, l1, l2, t->nodes[i]->data->num_adj,t->nodes[j]->data->num_adj );
        //printf("%d %d %d\n", i, j, est);
      }  
    }
  }
  fclose(fj2);

  for(i = 0; i < 2*num_genes + 1; i++)
    free(sel_adj[i]);
  free(sel_adj);  
}

void run_jk(tree t, double perc, int num_bs, double ***mlist)
{
  int i, j, b, l1, l2, est, sh_adj;
  int num_genes = t->nodes[1]->data->size_genome;
  int **sel_adj = (int **)calloc(2*num_genes+1,sizeof(int *));
  for(i = 0; i < 2*num_genes + 1; i++)
    sel_adj[i] = (int *)calloc(2*num_genes+1, sizeof(int));

  //FILE *fj2 = fopen(filename, "w"); //jackknife
  print_all_dist(t,mlist[0]);
  //printf("writing to %s\n", filename);
  printf("Computing bootstrap support values\n");
  for(b = 1; b <= num_bs; b++){
    //printf("bs %d\n", b);
    int num_adj = 0;
    unset_adjacencies(sel_adj, num_genes);
    for(i = 1; i <= t->num_leaves; i++)
      num_adj += set_adj_jk(sel_adj, t->nodes[i]);
    jackknife(sel_adj, int(double(num_adj)*perc), num_genes);
    //fprintf(fj2, "%d\n", t->num_leaves);
    for (i = 1; i <= t->num_leaves; i++){
      for (j = i+1; j <= t->num_leaves; j++){
        sh_adj = get_shared_adj_jk(t->nodes[i], t->nodes[j], sel_adj);
        l1 = num_lin_chrm(t->nodes[i]->data);
        l2 = num_lin_chrm(t->nodes[j]->data);
//printf("sh adj %d %d %d\n", sh_adj, l1, l2);
        est = GetEstimate(num_genes, l1, l2, sh_adj);
        if( (sh_adj >= t->nodes[i]->data->num_adj) || (sh_adj >= t->nodes[j]->data->num_adj) )
          est = 0;
        //fprintf(fj2, "%d %d %d\n", i, j, est);
        mlist[b][i-1][j-1] = est;
        //printf("sh adj %d %d %d %d %d\n", sh_adj, l1, l2, t->nodes[i]->data->num_adj,t->nodes[j]->data->num_adj );
        //printf("%d %d %d\n", i, j, est);
      }  
    }
  }
  //fclose(fj2);

  for(i = 0; i < 2*num_genes + 1; i++)
    free(sel_adj[i]);
  free(sel_adj);  
}

//JK

/////INPUT HANDLING FUNCTIONS
//IHF

tree read_leaves(tree t, char *filename, char **names, int num_leaves, int num_genes)
{

  int x = 1, i, prev, count, numadj, e1, e2;
  char c;
  char str[MAXNAMESIZE];


  genome gen;

  FILE *fp = fopen(filename, "r");
  count = 1;
  while(x != EOF){
    while((x = fscanf(fp, "%d", &i)) == 1){
      //printf("%d ",i);
      if(prev != -1){
        if (i < 0){
          //printf("%d %d\n", prev, -2*i);
	  e1 = prev;
	  e2 = -2*i;
          prev = -2*i - 1;
        }
        else if (i > 0){
          //printf("%d %d\n", prev, 2*i - 1);
	  e1 = prev;
	  e2 = 2*i - 1;
          prev = 2*i;
        }
      }
      else if (prev == -1){
        if (i < 0){
          //printf("%d %d\n", -2*i, -1);
	  e1 = -2*i;
	  e2 = -1;
          prev = -2*i - 1;
        }
        else if (i > 0){
          //printf("%d %d\n", 2*i - 1, -1);
	  e1 = 2*i-1;
	  e2 = -1;
          prev = 2*i;
        }
      }
      //printf("%d %d numadj = %d\n", e1, e2, numadj);
      gen->adjs[numadj] = (adjacency)malloc(sizeof(struct adj_struct));
      gen->adjs[numadj]->e1 = gen->elems[e1];
      gen->elems[e1]->adj = gen->adjs[numadj];
      if (e2 != -1){
        gen->adjs[numadj]->e2 = gen->elems[e2];
        gen->elems[e2]->adj = gen->adjs[numadj];
      }
      else
        gen->adjs[numadj]->e2 = NULL;
      numadj++;
    }
    if (x != EOF){
      x = 1;
      c = fgetc(fp);
      if (c == '>'){
	if (count > 1){
	  gen->size_genome = num_genes;
          gen->num_adj = numadj;
	}
        fscanf(fp, "%s", names[count-1]);//read species name
        while((c = fgetc(fp)) != '\n');//ignore rest of line
        //printf("%s allocating gen\n", str);
        t->nodes[count] = (node)malloc(sizeof(struct node_struct));
        gen = (genome)malloc(sizeof(struct genome_struct));
        gen->adjs = (adjacency *)malloc(2*num_genes*sizeof(adjacency));
        gen->elems = (element *)malloc((2*num_genes+1)*sizeof(element));
        t->nodes[count]->data = gen;
        gen->elems[0] = NULL;
        for(int j = 1; j < (2*num_genes + 1); j++){
          gen->elems[j] = (element)malloc(sizeof(struct elem_struct));
          gen->elems[j]->elem_code = j;
          gen->elems[j]->adj = NULL;
        }
        prev = -1;
	count++;
        //printf("numadj %d\n", numadj);
        numadj = 0;
      }
      else if (c == '$'){
        assert(prev != -1);
        //printf("adj: %d %d\n", prev, -1);
	e1 = prev;
	e2 = -1;
	gen->adjs[numadj] = (adjacency)malloc(sizeof(struct adj_struct));
        gen->adjs[numadj]->e1 = gen->elems[e1];
        gen->elems[e1]->adj = gen->adjs[numadj];
        gen->adjs[numadj]->e2 = NULL;
        prev = -1;      
	numadj++;
      }
    }
  }
  gen->size_genome = num_genes;
  gen->num_adj = numadj;

  fclose(fp);
  return t;
}

int check_format(char *filename, int *num_leaves, int *num_genes)
{
  int i, numgenes, prevnumgenes;
  int count, x;
  char c, str[MAXNAMESIZE];
  int state; // 

  FILE *fp = fopen(filename, "r");
  count = 0;
  c = fgetc(fp);
  if (c != '>'){
    printf("%s: format incorrect; file should begin with \">\"\n", filename);
    return -1;
  }
  else{
    fscanf(fp, "%s", str);
    if (strlen(str) < 1){
      printf("%s: format incorrect; species name %d must be at least one character long\n", filename, count);
      return -1;
    }      
    //printf("%s\n", str);
    while((c = fgetc(fp)) != '\n');
    count++;
    numgenes = prevnumgenes = 0;
  }
  x = 1;

  while(x != EOF){
    while((x = fscanf(fp, "%d", &i)) == 1){
      numgenes++;
      state = 0;                           
    }
    //printf("%d genes read\nx = %d\n", numgenes, x);
    if (x != EOF){
      x = 1;
      c = fgetc(fp);
      if (c == '>'){
        if (state != 1){
          printf("%s: format incorrect; chromosome should end with $\n", filename);
          return -1;
        }
        fscanf(fp, "%s", str);
        if (strlen(str) < 1){
          printf("%s: format incorrect; species name %d must be at least one character long\n", filename, count);
          return -1;
        }      
        //printf("%s\n", str);
        while((c = fgetc(fp)) != '\n');
	if ((count > 2) && (numgenes != prevnumgenes)){
          printf("%s: format incorrect; genomes have different number of genes\n", filename);
          return -1;
        }
	prevnumgenes = numgenes;
        numgenes = 0;
        count++;
      }
      else if (c == '$'){
        if (state != 0){
          printf("%s: format incorrect; chromosome should be nonempty and end with $\n", filename);
          return -1;
        }
        state = 1;
        //printf("$ found\n");
	//getchar();
      }
      else{
        printf("%s: format incorrect; unexpected character %c found\n", filename, c);
        return -1;
      }
    }
    else if (state == 0){
      printf("%s: format incorrect; chromosome should end with $\n", filename);
      return -1;
    }
  }
  if (numgenes == 0){
    printf("%s: format incorrect; no genes found\n", filename);
    return -1;
  }
  if ((count == 2) && (numgenes != prevnumgenes)){
    printf("%s: format incorrect; genomes have different number of genes\n", filename);
    return -1;
  }
  if (count < 2){
    printf("There must be at least 2 genomes, only %d found\n", count);
    return -1;
  }
  //printf("num of genes read = %d\n", numgenes);
  //printf("num of genomes read =  %d\n", count);
  *num_genes = numgenes;
  *num_leaves = count;
  fclose(fp);
  return 1;  
}

int main(int argc, char *argv[])
{
  if (argc != 5 && argc != 6){
    printf("usage: tiba <input filename> <output filename> <NJ|FME> <BC|PJ> [# bootstraps>]\ntree reconstruction methods: NJ, FME (case-sensitive), NJ: Neighbor-Joining, FME: FastME\nbootstrap methods: BC, PJ (case-sensitive) BC: Classic Bootstrap, PJ: Parsimony Jackknife\n");
    exit(1);
  }
 
  int num_genes, num_leaves, num_bs;
  if (argc == 6)
    num_bs = atoi(argv[5]);
  else
    num_bs = 100;
  char matrixfile[] = "matrices";
  char ofile1[100], ofile2[100];
  char method[5];

  strcpy(method, argv[3]);
  if ((strcmp("NJ", method) != 0) && (strcmp("FME", method) != 0)){
    printf("Tree reconstruction method must be either 'NJ' or 'FME'. Please try again\n");
    exit(1);
  }

  if (check_format(argv[1], &num_leaves, &num_genes) == -1)
    exit(1);
  

  char **names = (char **)malloc(sizeof(char *) * num_leaves);
  for(int i = 0; i < num_leaves; i++)
    names[i] = (char *)malloc(sizeof(char) * MAXNAMESIZE);

  tree t = (tree)malloc(sizeof(struct tree_struct));
  t->nodes = (node *)malloc(sizeof(node) * (1+num_leaves));
  t->num_leaves = num_leaves;
  t->root = NULL;
  for(int i = 0; i <= num_leaves; i++)
    t->nodes[i] = NULL;
  read_leaves(t, argv[1], names, num_leaves, num_genes);

  printf("Input file read: ");
  printf("#genes = %d ", num_genes);
  printf("#leaves = %d\n", num_leaves);
  //for(int i = 0; i < num_leaves; i++)
  //  printf("%s\n", names[i]);

  //printf("Genomes read\n");
  strcpy(ofile1, argv[2]);
  strcpy(ofile2, argv[2]);
  strcat(ofile1, "_1");
  strcat(ofile2, "_2");
 
/*
  if ((argc == 7) && (strcmp("-m",argv[6])==0)){
    printf("High Memory mode\n");
    double ***mlist = (double ***)malloc((1+num_bs)*sizeof(double **));
    for (int i = 0; i <= num_bs; i++){
      mlist[i] = (double **)malloc(num_leaves * sizeof(double *));
      for (int j = 0; j < num_leaves; j++)
        mlist[i][j] = (double *)malloc((num_leaves) * sizeof(double));
    } 
    if (strcmp("PJ",argv[4]) == 0)
      run_pj(t, num_bs, mlist);
    else if (strcmp("BC",argv[4]) == 0)
      run_bc(t, num_bs, mlist);
    else
      printf("Bootstrap method parameter must be PJ or BC. Please try again.\n");
    printf("Matrices computed\n");

    write_nj_tree_from_matrix(mlist, ofile1, ofile2, num_bs, num_leaves, num_genes);
    printf("Output written to:\n%s (newick format with branch lengths) and\n%s (newick format with support values)\n", ofile1, ofile2);
    for (int i = 0; i <= num_bs; i++){
      for (int j = 0; j < num_leaves; j++)
        free(mlist[i][j]);
      free(mlist[i]);
    } 
    free(mlist);
  }
*/
    if (strcmp("PJ",argv[4]) == 0)
      run_pj(matrixfile, t, num_bs);
    /*else if (strcmp("50JK",argv[3]) == 0)
      run_jk(matrixfile, t, 0.5, num_bs);
    else if (strcmp("37JK",argv[3]) == 0)
      run_jk(matrixfile, t, 0.37, num_bs);
    else if (strcmp("BP",argv[3]) == 0)
      run_bp(matrixfile, t, num_bs);*/
    else if (strcmp("BC",argv[4]) == 0)
      run_bc(matrixfile, t, num_bs);
    else{
      printf("Bootstrap method parameter must be BC or PJ. Please try again.\n");
      exit(1);
    }
    printf("Matrices written to file %s\n", matrixfile);
   
    if (strcmp("NJ", method) == 0){
      printf("NJ\n");    
      write_nj_tree(matrixfile, ofile1, ofile2, num_bs, num_leaves, num_genes, names);
    }
    else{
      char fmename[MAXNAMESIZE];
      FILE *fp = fopen("fmename", "r");
      fscanf(fp, "%s", fmename);
      fclose(fp);
      if (strlen(fmename) < 1){
        printf("Cannot read file \'fmename\' for FastME executable name. Please run \'configure\' and \'make\' again.\n");
	exit(1);
      }
      char dm[] = "fmedm";
      generate_fme_matrix(matrixfile, dm, num_bs, num_leaves, num_genes);
      char cmd[100];
      sprintf(cmd, "%s -i fmedm --datasets=%d -o fmetree\0", fmename, num_bs+1);
      system(cmd);
      printf("FME\n");    
      char fmetree[] = "fmetree";
      write_fme_tree(fmetree, ofile1, ofile2, num_bs, num_leaves, num_genes, names);
    }
    printf("Output written to:\n%s (newick format with branch lengths) and\n%s (newick format with support values)\n", ofile1, ofile2);

  destroy_tree(t);
  return 0;
}
//IHF


